/*
 * intergrated input support
 */
#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "../port/error.h"

/* one line per event, the type followed by tab seperated fields of values:
  * pen	abs x y z	til x y but b
  * mse	abs x y z	rot r	whl x	but b
  * mse	rel x y	but b
  * key	key k	val d 
  */

/* or more broken down events:
  * pen	x	y	z	pressure	tiltx	tilty
  * mouse	x	y	z	r
  * key	k	v
  * where buttons and wheels are special sorts of keys
  */

enum {
	Qdir,
	Qinput,
	Qin,
	Qmap,
};

static Dirtab inputdir[] = {
	".",	{Qdir, 0, QTDIR},	0,	DMDIR|0555,
	"input",	{Qinput},		0,	0440,
	"inputin",	{Qin},		0,	0220,
	"inputmap",	{Qmap},		0,	0660,
};

static void
inputinit(void)
{
}

static Chan*
inputattach(char *param)
{
	return devattach(L'ι', param);
}

static Walkqid*
inputwalk(Chan *c, Chan *nc, char **name, int nname)
{
	return devwalk(c, nc, name, nname, inputdir, nelem(inputdir), devgen);
}

static int
inputstat(Chan *c, uchar *db, int n)
{
	return devstat(c, db, n, inputdir, nelem(inputdir), devgen);
}

static Chan*
inputopen(Chan *c, int omode)
{
	switch((ulong)c->qid.path){
	default:
		error(Eperm);
	case Qdir:
		if(omode != OREAD)
			error(Eperm);
		break;
	case Qinput:
		if(omode != OREAD)
			error(Eperm);
		break;
	case Qin;
		if(omode != OWRITE)
			error(Eperm);
		break;
	case Qmap;
		break;
	}
	c = devopen(c, omode, inputdir, nelem(inputdir), devgen);
	c->mode = openmode(omode);
	c->flag |= COPEN;
	c->offset = 0;
	return c;
}

static void
inputclose(Chan *)
{
}

static long
inputread(Chan *c, void *v, long n, vlong off)
{
	switch((ulong)c->qid.path){
	default:
		error(Eperm);
	case Qdir:
		return devdirread(c, v, n, inputdir, nelem(inputdir), devgen);
	case Qinput:
		return inputread(v, n, off);
	case Qin:
		error(Eperm);
	case Qmap:
		return mapread(v, n, off);
	}
}

static long
inputwrite(Chan *c, void *v, long n, vlong off)
{
	switch((ulong)c->qid.path){
	default:
		error(Eperm);
	case Qinput:
		error(Eperm);
	case Qin:
		return inputwrite(v, n, off);
	case Qmap:
		return mapwrite(v, n, off);
	}
}

Dev inputdevtab = {
	L'ι',
	"in",

	devreset,
	inputinit,
	devshutdown,
	inputattach,
	inputwalk,
	inputstat,
	inputopen,
	devcreate,
	inputclose,
	inputread,
	devbread,
	inputwrite,
	devbwrite,
	devremove,
	devwstat,
};
